/*
 * File:   newmain.c
 * Author: Joe
 *
 * Created on July 1, 2018, 11:32 AM
 */

// PIC16LF18855 Configuration Bit Settings

// 'C' source line config statements

// CONFIG1
#pragma config FEXTOSC = OFF    // External Oscillator mode selection bits (EC above 8MHz; PFM set to high power)
//#pragma config RSTOSC = EXT1X   // Power-up default value for COSC bits (EXTOSC operating per FEXTOSC bits)
#pragma config CLKOUTEN = OFF   // Clock Out Enable bit (CLKOUT function is disabled; i/o or oscillator function on OSC2)
#pragma config CSWEN = ON       // Clock Switch Enable bit (Writing to NOSC and NDIV is allowed)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable bit (FSCM timer enabled)

// CONFIG2
#pragma config MCLRE = ON       // Master Clear Enable bit (MCLR pin is Master Clear function)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config LPBOREN = OFF    // Low-Power BOR enable bit (ULPBOR disabled)
#pragma config BOREN = OFF       // Brown-out reset enable bits (Brown-out Reset Enabled, SBOREN bit is ignored)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (VBOR) set to 1.9V on LF, and 2.45V on F Devices)
#pragma config ZCD = OFF        // Zero-cross detect disable (Zero-cross detect circuit is disabled at POR.)
#pragma config PPS1WAY = ON     // Peripheral Pin Select one-way control (The PPSLOCK bit can be cleared and set only once in software)
#pragma config STVREN = OFF      // Stack Overflow/Underflow Reset Enable bit (Stack Overflow or Underflow will cause a reset)

// CONFIG3
#pragma config WDTCPS = WDTCPS_31// WDT Period Select bits (Divider ratio 1:65536; software control of WDTPS)
#pragma config WDTE = OFF        // WDT operating mode (WDT enabled regardless of sleep; SWDTEN ignored)
#pragma config WDTCWS = WDTCWS_7// WDT Window Select bits (window always open (100%); software control; keyed access not required)
#pragma config WDTCCS = SC      // WDT input clock selector (Software Control)

// CONFIG4
#pragma config WRT = OFF        // UserNVM self-write protection bits (Write protection off)
#pragma config SCANE = available// Scanner Enable bit (Scanner module is available for use)
#pragma config LVP = ON         // Low Voltage Programming Enable bit (Low Voltage programming enabled. MCLR/Vpp pin function is MCLR.)

// CONFIG5
#pragma config CP = OFF         // UserNVM Program memory code protection bit (Program Memory code protection disabled)
#pragma config CPD = OFF        // DataNVM code protection bit (Data EEPROM code protection disabled)

/* Setup Chip Frequency and Delay Timer */
#pragma config RSTOSC = HFINT1  // Set up Startup Speed (1Mhz)
#define _XTAL_FREQ 1000000      // The delay function uses the _XTAL_FREQ to calculate it's timings etc, but unless the physical clock is set correctly then it won't work properly.

/* Local Defines */


/* #define RESET RE3 */
#define KeyFlag         RA0     // QH' Pin of Shift Register 1
#define RCLK1           RA1     // Shift Register 1 - RCLK   (Status Byte)
#define RCLK0           RA2     // Shift Register 0 - RCLK  (Key Byte)
#define SRCLK           RA3     // Shared SRCLK
#define SER             RA4     // Shared SER
#define CapLock         RA5     // CapsLock On/Off

#define Keyboard_Row1   RA7     // Keyboard Row 1
#define Keyboard_Row2   RA6     // Keyboard Row 2
#define Keyboard_Row3   RC0     // Keyboard Row 3
#define Keyboard_Row4   RC1     // Keyboard Row 4
#define Keyboard_Row5   RC2     // Keyboard Row 5

#define Keyboard_Col1   RB7     // Keyboard Column 1
#define Keyboard_Col2   RB6     // Keyboard Column 2
#define Keyboard_Col3   RB5     // Keyboard Column 3
#define Keyboard_Col4   RB4     // Keyboard Column 4
#define Keyboard_Col5   RB3     // Keyboard Column 5
#define Keyboard_Col6   RB2     // Keyboard Column 6
#define Keyboard_Col7   RB1     // Keyboard Column 7
#define Keyboard_Col8   RB0     // Keyboard Column 8

#define Keyboard_Col9   RC7     // Keyboard Column 9
#define Keyboard_Col10  RC6     // Keyboard Column 10
#define Keyboard_Col11  RC5     // Keyboard Column 11
#define Keyboard_Col12  RC4     // Keyboard Column 12
#define Keyboard_Col13  RC3     // Keyboard Column 13

#define Keyboard_Row1_DIR TRISA7    // Row input/output directions
#define Keyboard_Row2_DIR TRISA6
#define Keyboard_Row3_DIR TRISC0
#define Keyboard_Row4_DIR TRISC1
#define Keyboard_Row5_DIR TRISC2

typedef struct {unsigned b0:1,b1:1,b2:1,b3:1,b4:1,b5:1,b6:1,b7:1;}var_bits;
#define bit_of(var,num)  ((  (var_bits*)&var  )->b##num)

#define Keybit0 bit_of(Key,0)
#define Keybit1 bit_of(Key,1)
#define Keybit2 bit_of(Key,2)
#define Keybit3 bit_of(Key,3)
#define Keybit4 bit_of(Key,4)
#define Keybit5 bit_of(Key,5)
#define Keybit6 bit_of(Key,6)
#define Keybit7 bit_of(Key,7)
  
#define CapsKey bit_of(Statuses,0)     // Caps Lock
#define ShiftKey bit_of(Statuses,1)    // Shift
#define CtrlKey bit_of(Statuses,2)     // Control
#define AltKey bit_of(Statuses,3)      // Alt
#define FnKey bit_of(Statuses,4)       // Fn
#define WinKey bit_of(Statuses,5)      // Win Key
#define StatBit6 bit_of(Statuses,6)    // ** reserved **
#define KeyReady bit_of(Statuses,7)    // Char In Buffer

unsigned char Key, LastKey, LastCaps, RepeatTime, CapsLock, Statuses = 0x00;
unsigned char Read_Key(unsigned char row, unsigned char column);

void InitPIC(void);
void KeyOut(unsigned char Key);
void StatusOut(unsigned char Statuses);
void ClockOut(unsigned char KeyByte);
void Wait50(void);
void Wait10(void);

#include <xc.h>

int main() {
    // Set up the PIC's pins, registers, etc.
    // Initialize Port A
    PORTA = 0;          // Set Port to Zero
    LATA = 0;           // Set Latch to Zero
    ANSELA = 0;         // Turn off analog
    TRISA = 0b00000001; // Set input/output
    INLVLA = 0;         // Set input level to TTL trigger
    
// Initialize Port B
    PORTB = 0;          // Set Port to Zero
    LATB = 0;           // Set Latch to Zero
    ANSELB = 0;         // Turn off analog
    TRISB = 0b11111111; // Set input/output
    INLVLB = 0;         // Set input level to TTL trigger

// Initialize Port C
    PORTC = 0;          // Set Port to Zero
    LATC = 0;           // Set Latch to Zero
    ANSELC = 0;         // Turn off analog
    TRISC = 0b11111000; // Set input/output

    SRCLK=0;            // Pull Down SRCLK
    RCLK0=1;            // Pull Up RCLK0
    RCLK1=1;            // Pull Up RCLK1
    
    
    while(1)    // Main Loop
      {
        if (KeyFlag==1){continue;}  // If key is in the buffer and hasn't been read out by JCM-1 yet, then spinlock.
          else {                    // If key has been read, we can scan the keyboard again.
        
          // read and set special keys key-down status before checking any other keys.
          Statuses=0;             // clear modifiers on every scan
          CapsKey=Read_Key(3,1);  // read caps lock, and set status if pressed
          ShiftKey=Read_Key(4,1); // read shift, and set status if pressed
          CtrlKey=Read_Key(5,1);  // read control, and set status if pressed
          AltKey=Read_Key(5,3);   // read Alt, and set status if pressed
          FnKey=Read_Key(5,5);    // read Fn, and set status if pressed
          WinKey=Read_Key(5,2);   // read Win, and set status if pressed
          //StatusOut(Statuses);    // save Statuses to '595.
       
          // Deal With Caps Lock
            if (Read_Key(3,1)==1){
                Wait10();             // Deal with Keybounce
                if (Read_Key(3,1)==1){
                    // if (LastCaps==1){ key was down on last loop, do nothing }
                    if (LastCaps==0){   // if key was up on last loop...
                        LastCaps=1;             // Set the flag so we can check on next loop if the key is still pressed.
                        CapLock = ~CapLock;     // Toggle the CapLock flag.
                    }
                }
            }
            if (Read_Key(3,1)==0 && LastCaps==1){LastCaps=0;}   // If Caps Lock isn't pressed, but was down on last loop
                                                                // that means we've successfully toggled the CapLock flag.
                                                                // Set LastCaps to 0 so as to reset the logic above.

          
           /*
            How to read it using a genericized loop
            int i,j;
            char a;
            for (i = 1; i < 5; i++)
            {
                for (j = 1; i < 13; i++)
                {
                    if (Read_Key(i,j))
                    {
                        Wait10();
                        if(Read_Key(i,j)){
                            Key=array(ShiftKey,CapLock,CtrlKey,FnKey,WinKey,i,j)
                        }
                    }
                    
                }
            }
           */
          
         // Read Row 1
          Key=0;
               if(Read_Key(1,1)==1)   { Wait10();  if(Read_Key(1,1)==1){ Key=27; } } // Escape
          else if(Read_Key(1,2)==1)   { Wait10();  if(Read_Key(1,2)==1){(ShiftKey) ? (Key='!') : (Key='1');}}
          else if(Read_Key(1,3)==1)   { Wait10();  if(Read_Key(1,3)==1){(ShiftKey) ? (Key='@') : (Key='2');}}  
          else if(Read_Key(1,4)==1)   { Wait10();  if(Read_Key(1,4)==1){(ShiftKey) ? (Key='#') : (Key='3');}}  
          else if(Read_Key(1,5)==1)   { Wait10();  if(Read_Key(1,5)==1){(ShiftKey) ? (Key='$') : (Key='4');}}  
          else if(Read_Key(1,6)==1)   { Wait10();  if(Read_Key(1,6)==1){(ShiftKey) ? (Key='%') : (Key='5');}}  
          else if(Read_Key(1,7)==1)   { Wait10();  if(Read_Key(1,7)==1){(ShiftKey) ? (Key='^') : (Key='6');}}  
          else if(Read_Key(1,8)==1)   { Wait10();  if(Read_Key(1,8)==1){(ShiftKey) ? (Key='&') : (Key='7');}}  
          else if(Read_Key(1,9)==1)   { Wait10();  if(Read_Key(1,9)==1){(ShiftKey) ? (Key='*') : (Key='8');}}  
          else if(Read_Key(1,10)==1)  { Wait10();  if(Read_Key(1,10)==1){(ShiftKey) ? (Key='(') : (Key='9');}}  
          else if(Read_Key(1,11)==1)  { Wait10();  if(Read_Key(1,11)==1){(ShiftKey) ? (Key=')') : (Key='0');}}  
          else if(Read_Key(1,12)==1)  { Wait10();  if(Read_Key(1,12)==1){(ShiftKey) ? (Key='_') : (Key='-');}}
          else if(Read_Key(1,13)==1)  { Wait10();  if(Read_Key(1,13)==1){(ShiftKey) ? (Key='+') : (Key='=');}} 

          
          // Read Row 2
          else if(Read_Key(2,1)==1)   { Wait10();  if(Read_Key(2,1)==1){ Key=9; } }  // Tab
          else if(Read_Key(2,2)==1)   { Wait10();  if(Read_Key(2,2)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='Q';} else if (CtrlKey){Key=17;} else {Key='q';}} }
          else if(Read_Key(2,3)==1)   { Wait10();  if(Read_Key(2,3)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='W';} else if (CtrlKey){Key=23;} else {Key='w';}} }
          else if(Read_Key(2,4)==1)   { Wait10();  if(Read_Key(2,4)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='E';} else if (CtrlKey){Key=5;} else {Key='e';}} }
          else if(Read_Key(2,5)==1)   { Wait10();  if(Read_Key(2,5)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='R';} else if (CtrlKey){Key=18;} else {Key='r';}} }
          else if(Read_Key(2,6)==1)   { Wait10();  if(Read_Key(2,6)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='T';} else if (CtrlKey){Key=20;} else {Key='t';}} }
          else if(Read_Key(2,7)==1)   { Wait10();  if(Read_Key(2,7)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='Y';} else if (CtrlKey){Key=25;} else {Key='y';}} }
          else if(Read_Key(2,8)==1)   { Wait10();  if(Read_Key(2,8)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='U';} else if (CtrlKey){Key=21;} else {Key='u';}} }
          else if(Read_Key(2,9)==1)   { Wait10();  if(Read_Key(2,9)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='I';} else if (CtrlKey){Key=9;} else {Key='i';}} }
          else if(Read_Key(2,10)==1)  { Wait10();  if(Read_Key(2,10)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='O';} else if (CtrlKey){Key=15;} else {Key='o';}} }
          else if(Read_Key(2,11)==1)  { Wait10();  if(Read_Key(2,11)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='P';} else if (CtrlKey){Key=16;} else {Key='p';}} }
          else if(Read_Key(2,12)==1)  { Wait10();  if(Read_Key(2,12)==1){if(ShiftKey){Key='{';} else {Key='[';} } }
          else if(Read_Key(2,13)==1)  { Wait10();  if(Read_Key(2,13)==1){if(ShiftKey){Key='}';} else {Key=']';} } }  

          // Read Row 3
          // else if(Read_Key(3,1)==1){ Wait10();  if(Read_Key(5,4)==1){Key=32; }} /* Caps Lock */
          else if(Read_Key(3,2)==1)   { Wait10();  if(Read_Key(3,2)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='A';} else if (CtrlKey){Key=1;} else {Key='a';} }}
          else if(Read_Key(3,3)==1)   { Wait10();  if(Read_Key(3,3)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='S';} else if (CtrlKey){Key=19;} else {Key='s';} }}
          else if(Read_Key(3,4)==1)   { Wait10();  if(Read_Key(3,4)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='D';} else if (CtrlKey){Key=4;} else {Key='d';} }}
          else if(Read_Key(3,5)==1)   { Wait10();  if(Read_Key(3,5)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='F';} else if (CtrlKey){Key=6;} else {Key='f';} }}
          else if(Read_Key(3,6)==1)   { Wait10();  if(Read_Key(3,6)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='G';} else if (CtrlKey){Key=7;} else {Key='g';} }}
          else if(Read_Key(3,7)==1)   { Wait10();  if(Read_Key(3,7)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='H';} else if (CtrlKey){Key=8;} else {Key='h';} }}
          else if(Read_Key(3,8)==1)   { Wait10();  if(Read_Key(3,8)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='J';} else if (CtrlKey){Key=10;} else {Key='j';} }}
          else if(Read_Key(3,9)==1)   { Wait10();  if(Read_Key(3,9)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='K';} else if (CtrlKey){Key=11;} else {Key='k';} }}
          else if(Read_Key(3,10)==1)  { Wait10();  if(Read_Key(3,10)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='L';} else if (CtrlKey){Key=12;} else {Key='l';} }}
          else if(Read_Key(3,11)==1)  { Wait10();  if(Read_Key(3,11)==1){if(ShiftKey){Key=':';} else {Key=';';} }} 
          else if(Read_Key(3,12)==1)  { Wait10();  if(Read_Key(3,12)==1){if(ShiftKey){Key='"';} else {Key='\'';} }}
          else if(Read_Key(3,13)==1)  { Wait10();  if(Read_Key(3,13)==1){Key=13; }}     // Carriage Return

          // Read Row 4        
          // else if(Read_Key(4,1)==1)   { Wait10();  if(Read_Key(4,1)==1){Key=32; }} /* Shift */
          else if(Read_Key(4,2)==1)   { Wait10();  if(Read_Key(4,2)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='Z';} else if (CtrlKey){Key=26;} else {Key='z';} }}
          else if(Read_Key(4,3)==1)   { Wait10();  if(Read_Key(4,3)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='X';} else if (CtrlKey){Key=24;} else {Key='x';} }}
          else if(Read_Key(4,4)==1)   { Wait10();  if(Read_Key(4,4)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='C';} else if (CtrlKey){Key=3;} else {Key='c';} }}
          else if(Read_Key(4,5)==1)   { Wait10();  if(Read_Key(4,5)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='V';} else if (CtrlKey){Key=22;} else {Key='v';} }}
          else if(Read_Key(4,6)==1)   { Wait10();  if(Read_Key(4,6)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='B';} else if (CtrlKey){Key=2;} else {Key='b';} }}
          else if(Read_Key(4,7)==1)   { Wait10();  if(Read_Key(4,7)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='N';} else if (CtrlKey){Key=14;} else {Key='n';} }}
          else if(Read_Key(4,8)==1)   { Wait10();  if(Read_Key(4,8)==1){if((ShiftKey || CapLock)&&!CtrlKey){Key='M';} else if (CtrlKey){Key=13;} else {Key='m';} }}
          else if(Read_Key(4,9)==1)   { Wait10();  if(Read_Key(4,9)==1){if(ShiftKey){Key='<';} else {Key=',';} }}
          else if(Read_Key(4,10)==1)  { Wait10();  if(Read_Key(4,10)==1){if(ShiftKey){Key='>';} else {Key='.';} }}
          else if(Read_Key(4,11)==1)  { Wait10();  if(Read_Key(4,11)==1){if(ShiftKey){Key='?';} else {Key='/';} } }
          else if(Read_Key(4,12)==1)  { Wait10();  if(Read_Key(4,12)==1){Key=193; }}    // up arrow
          else if(Read_Key(4,13)==1)  { Wait10();  if(Read_Key(4,13)==1){Key=194; }}    // PgDn (masquerading as down arrow)

          // Read Row 5
          //else if(Read_Key(5,1)==1) { Wait10();  if(Read_Key(5,4)==1){Key=32; }} /* Control */
          //else if(Read_Key(5,2)==1) { Wait10();  if(Read_Key(5,4)==1){Key=32; }} /* Win Key */
          //else if(Read_Key(5,3)==1) { Wait10();  if(Read_Key(5,4)==1){Key=32; }} /* Alternate */
          else if(Read_Key(5,4)==1)   { Wait10();  if(Read_Key(5,4)==1){Key=32; }}
          //else if(Read_Key(5,5)==1) { Wait10();  if(Read_Key(5,4)==1){Key=32; }} /* Fn Key */
          else if(Read_Key(5,6)==1)   { Wait10();  if(Read_Key(5,6)==1){Key=127; }}     // Backspace
          else if(Read_Key(5,7)==1)   { Wait10();  if(Read_Key(5,7)==1){if(ShiftKey){Key='~';} else {Key='`';} }}
          else if(Read_Key(5,8)==1)   { Wait10();  if(Read_Key(5,8)==1){if(ShiftKey){Key='|';} else {Key='\\';} }}
          else if(Read_Key(5,9)==1)   { Wait10();  if(Read_Key(5,9)==1){Key=127; }}      // Delete
          else if(Read_Key(5,10)==1)  { Wait10();  if(Read_Key(5,10)==1){Key=196; }}    // left arrow
          else if(Read_Key(5,11)==1)  { Wait10();  if(Read_Key(5,11)==1){Key=194; }}    // down arrow
          else if(Read_Key(5,12)==1)  { Wait10();  if(Read_Key(5,12)==1){Key=195; }}    // right arrow
          else if(Read_Key(5,13)==1)  { Wait10();  if(Read_Key(5,13)==1){Key=195; }}    // PgUp (masquerading as up arrow)
          else {Key=0;LastKey=0;RepeatTime=0;}

          if(Key!=0){   // check for key repeat
              if (Key==LastKey){        // if the new key is the same as the last key, and ...
                  if (RepeatTime==40){      // if we've waited for a sufficient amount of time
                     Wait50();                  // then go into repeat mode
                     KeyOut(Key);
                  }
                  else {                    // we've not waited long enough, go into wait mode
                       RepeatTime++;            // increment the number of times we've waited
                       KeyOut(0);               // then zero key/stat so JCM-1 won't do anything AND so the microcontroller doesn't go into spinlock.
                  }
              }
              else {                    // if the new key is not the same as the last key, then..
                  RepeatTime=0;             // Reset the Repeat Timer
                  KeyOut(Key);              // send the new key out
                  LastKey=Key;              // save the key we just sent for later comparison.
              }
          }
        }
    }
} 

void KeyOut(unsigned char Key){
    RCLK0=0;
    ClockOut (Key);
    RCLK0=1;
    if (Key!=0){Statuses=Statuses+128;}
    StatusOut(Statuses);
}

void StatusOut(unsigned char Statuses){
    RCLK1=0;
    ClockOut (Statuses);
    RCLK1=1;
    return;
}

void ClockOut(unsigned char KeyByte){
    int bit;
    for( bit=7; bit >=0 ; bit-- ){
        SER= (KeyByte >> bit) & 1U;
        SRCLK=1;
        SRCLK=0;
   }
}

void Wait10(){  __delay_ms(3); }
void Wait50(){  __delay_ms(50); }

unsigned char Read_Key(unsigned char row, unsigned char column)
{
    Keyboard_Row1_DIR=1;
    Keyboard_Row2_DIR=1;
    Keyboard_Row3_DIR=1;
    Keyboard_Row4_DIR=1;
    Keyboard_Row5_DIR=1;

    switch(row)
    {
        case 1: //Row 1 selected
            Keyboard_Row1_DIR=0;
            Keyboard_Row1=1;
            break;
        case 2: //Row 2 selected
            Keyboard_Row2_DIR=0;
            Keyboard_Row2=1;
            break;
        case 3: //Row 3 selected
            Keyboard_Row3_DIR=0;
            Keyboard_Row3=1;
            break;
        case 4: //Row 4 selected
            Keyboard_Row4_DIR=0;
            Keyboard_Row4=1;
            break;
        case 5: //Row 5 selected
            Keyboard_Row5_DIR=0;
            Keyboard_Row5=1;
            break;
    }

    switch(column)
    {
        case 1:
            return Keyboard_Col1; break;
        case 2:
            return Keyboard_Col2; break;
        case 3:
            return Keyboard_Col3; break;
        case 4:
            return Keyboard_Col4; break;
        case 5:
            return Keyboard_Col5; break;
        case 6:
            return Keyboard_Col6; break;
        case 7:
            return Keyboard_Col7; break;
        case 8:
            return Keyboard_Col8; break;
        case 9:
            return Keyboard_Col9; break;
        case 10:
            return Keyboard_Col10; break;
        case 11:
            return Keyboard_Col11; break;
        case 12:
            return Keyboard_Col12; break;
        case 13:
            return Keyboard_Col13; break;
    }
return 0;
}